/*************************************************** */
/* Rule Set Based Access Control                     */
/* Author and (c) 1999-2009: Amon Ott <ao@rsbac.org> */
/* Memory allocation                                 */
/* Last modified: 30/Mar/2009                        */
/*************************************************** */

#ifndef __RSBAC_RKMEM_H
#define __RSBAC_RKMEM_H

#include <linux/init.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <linux/timer.h>

#ifdef CONFIG_RSBAC_INIT_DELAY
void rsbac_kmem_cache_sizes_init(void);
#else
void __init rsbac_kmem_cache_sizes_init(void);
#endif

#ifdef CONFIG_MMU
#define RSBAC_MAX_KMALLOC 131072
#else
#ifndef CONFIG_LARGE_ALLOCS
#define RSBAC_MAX_KMALLOC 1048576
#else
#define RSBAC_MAX_KMALLOC 33554432
#endif
#endif

#define rsbac_vmalloc(x) kmalloc(x, GFP_ATOMIC)
#define rsbac_vfree(x) kfree(x)

struct rsbac_delayed_kfree_t
  {
    struct timer_list timer;
    void * objp;
  };

/* alloc mem spinlock safe with GFP_ATOMIC */
extern void * rsbac_kmalloc (size_t size);

/* alloc outside locks with GFP_KERNEL */
extern void * rsbac_kmalloc_unlocked (size_t size);

/* Allocate memory. This function decides which alloc type (kmalloc or vmalloc) */
/* to use and returns the decision as boolean value (TRUE = vmalloc), */
/* if vmalloc_used_p is not NULL (what it should rather be...) */
extern void * rsbac_vkmalloc (size_t size, rsbac_boolean_t * vmalloc_used_p);

extern void * rsbac_vkmalloc_unlocked (size_t size, rsbac_boolean_t * vmalloc_used_p);

/**
 * rsbac_kfree - free previously allocated memory
 * @objp: pointer returned by kmalloc.
 *
 * Don't free memory not originally allocated by rsbac_kmalloc()
 * or you will run into trouble.
 *
 * We simply call general kfree, which does everything
 */
static inline void rsbac_kfree (const void *objp)
  {
    kfree(objp);
  }

/* Free the allocated mem. Needs old returned mem type value, */
/* or FALSE, if no type stored (kmalloc type assumed) */
static inline void rsbac_vkfree(void *objp, rsbac_boolean_t vmalloc_used)
  {
    if(vmalloc_used)
      vfree(objp);
    else
      kfree(objp);
  }

#endif
